//Dependencies
import * as validate from "./validators.js";

/*
 * Panda JS
*/

class Builder {


    //Constructor
    constructor(slides, userOptions){
        //Build our default options
        this.slides = slides;
        //Build our default options
        this.options = {
            ...this.buildDefaultOptions(userOptions),
            ...userOptions
        };
        this.resolver = '';
        this.results = {
            complete: false,
            inputs: []
        };
    }


    /**
     * Create a new panda pop
     *
     * @return {Promise}
    */
    async create(){
        //Promise
        return new Promise(async (resolve, reject) => {
            this.resolver = resolve;
            //Build it
            this.panda = await this.build();
            //Add our slides
            for(let [key, slide] of Object.entries(this.slides)) await this.addSlide(key, slide);
            //Configure the options
            await this.configureOptions();
            //Insert on the page
            document.querySelector('body').append(this.panda);
            //Open
            await this.open();
            //Either await the bindings or return depending on option
            if(this.options.return){
                this.bindings();
                return this.resolver(this.panda);
            } else {
                await this.bindings()
            }
        });
    }


    /**
     * Build the default options
     *
     * @param {Object} userOptions The user options for any overwrites
     * @return {Object} An object of options
    */
    buildDefaultOptions(userOptions){
        let options = {
            closeBtn:true,
            position:'center',
            return: false
        }
        //Return
        return options;
    }


    /**
     * Build the panda
     *
     * @return {Promise} HTMLElement The new panda
    */
    async build(){
        //Build the new Select HTML element and add it directly after the target
        let wrapper = document.createElement('div'),
            container = document.createElement('div'),
            pandaid = (this.options.id ? this.options.id : `pandajs-${Math.random().toString(36).substr(2, 4)}`);
        //Set attributes
        container.classList.add('container');
        wrapper.classList.add('pandajs');
        wrapper.setAttribute('id', pandaid);
        //Position
        wrapper.classList.add(this.options.position);
        //Close Btn?
        if(this.options.closeBtn){
            let btn = document.createElement('span');
            btn.classList.add('close');
            btn.innerHTML = await this.getIconContent('close');
            container.append(btn);
        }
        //Populate
        wrapper.appendChild(container);
        //Return
        return wrapper;
    }


    /**
     * Configure options for the panda
     *
     * @return {Promise}
    */
     async configureOptions(){

        //Return
        return;
    }


    /**
     * Open the panda
     *
     * @return {Promise}
    */
    async open(){
        //Open it (with a javascript animation to avoid css transform issues)
        let container = this.panda.querySelector('.container');
        await container.animate(
            [
                { transform: 'scale(0)' },
                { transform: 'scale(1.2)' },
                { transform: 'scale(1)' }
            ],
            {
                duration: 150,
                iterations: 1
            }
        );
        //Return
        return;
    }


    /**
     * Add a slide
     *
     * @param {Object} key - The key
     * @param {Object} slide - A slide object
     * @returns {Promise}
    */
    async addSlide(key, slide){
        //Slide wrapper and content
        let wrapper = document.createElement('div'),
            header = document.createElement('div'),
            body = document.createElement('div'),
            footer = document.createElement('div'),
            [title, message, inputs, buttons] = await Promise.all([
                this.buildTitle(slide),
                this.buildMessage(slide),
                this.buildFields(slide),
                this.buildButtons(slide)
            ]);
        //Build the sections
        if(title) header.append(title);
        if(message) body.append(message);
        if(inputs) body.append(inputs);
        if(buttons) footer.append(buttons);
        //Attributes
        header.classList.add('header-wrapper');
        body.classList.add('body-wrapper');
        footer.classList.add('footer-wrapper');
        //Append the sections to the slide
        if(header.hasChildNodes()) wrapper.append(header);
        if(body.hasChildNodes()) wrapper.append(body);
        if(footer.hasChildNodes()) wrapper.append(footer);
        //Attributes
        wrapper.classList.add('slide');
        wrapper.setAttribute('data-slide', parseInt(key) + 1);
        //Is this the primary one?
        if(key == 0) wrapper.classList.add('active');
        //Add it to the instance
        this.panda.querySelector('.container').append(wrapper);
        //Build the content
        return;
    }


    /* ------------------------------------------------------------------------------------------------------------------------
        HTML Builders
    ------------------------------------------------------------------------------------------------------------------------ */


    /**
     * Build the title
     *
     * @param {Object} slide - The slide we're building the title for
     * @returns {Promise}
    */
    async buildTitle(slide){
        //Undefined?
        if(slide.title == undefined || slide.title == '') return false;
        //Create
        let title = document.createElement('h3');
        //Set attributes and content
        title.innerHTML = slide.title;
        //Return
        return title;
    }


    /**
     * Build the message
     *
     * @param {Object} slide - The slide we're building the title for
     * @returns {Promise}
    */
    async buildMessage(slide){
        //Undefined?
        if(slide.message == undefined || slide.message == '') return false;
        //Create
        let message = document.createElement('p');
        //Set attributes and content
        message.innerHTML = slide.message;
        //Return
        return message;
    }


    /**
     * Build the buttons
     *
     * @param {Object} slide - The slide we're building the title for
     * @returns {Promise}
    */
    async buildButtons(slide){
        //Undefined?
        if(slide.buttons == undefined || slide.buttons == '' || slide.buttons.length == 0) return false;
        //Create
        let wrapper = document.createElement('div');
        //Set attributes and content
        wrapper.classList.add('controls-wrapper');
        //Loop the buttons
        for(let btn of slide.buttons){
            let button = document.createElement('button');
            //Build id
            let btnId = `btn-${Math.random().toString(36).substr(2, 4)}`
            //Custom?
            if(btn.action == 'custom'){
                let customAction = btn.customAction;
                if(typeof customAction != 'function') continue;
            }
            //Populate
            btn.id = btnId;
            button.innerHTML = btn.text;
            button.setAttribute('data-action', btn.action);
            button.setAttribute('data-btn-id', btnId);
            if(!['', undefined].includes(btn.class)) button.classList.add(btn.class);
            if(!['', undefined].includes(btn.id)) button.setAttribute('id', btn.id);
            //Append it
            wrapper.appendChild(button);
        }
        //Return
        return wrapper;
    }


    /**
     * Build the inputs
     *
     * @param {Object} slide - The slide we're building the inputs for
     * @returns {Promise}
    */
    async buildFields(slide){
        //Undefined?
        if(slide.inputs == undefined) return false;
        //Create wrapper
        let wrapper = document.createElement('div');
        //Set attributes and content
        wrapper.classList.add('inputs-wrapper');
        //Loop the buttons
        for(let input of slide.inputs){
            switch(input.type){
                case 'text':
                case 'email':
                case 'number':
                    wrapper.append(await this.buildInput(input));
                break;
                case 'select':
                    wrapper.append(await this.buildSelect(input));
                break;
            }
        }
        //Return
        return wrapper;
    }


    /* ------------------------------------------------------------------------------------------------------------------------
        Input types
    ------------------------------------------------------------------------------------------------------------------------ */


    /**
     * General input (text, email, number, etc)
     *
     * @param {Object} input - The input item
     * @returns {Promise}
    */
     async buildInput(input){
        let field = document.createElement('input');
        //Type
        field.setAttribute('type', input.type);
        //Set attributes
        if(!['', undefined].includes(input.class)) field.classList.add(input.class);
        if(!['', undefined].includes(input.id)) field.setAttribute('id', input.id);
        if(!['', undefined].includes(input.placeholder)) field.setAttribute('placeholder', input.placeholder);
        if(!['', undefined].includes(input.name)) field.setAttribute('name', input.name);
        if(!['', undefined].includes(input.value)) field.value = input.value;
        if(input.required === true) field.setAttribute('required', true);
        //Return
        return field;
    }


    /**
     * Select box
     *
     * @param {Object} input - The input item
     * @returns {Promise}
    */
    async buildSelect(input){
        let field = document.createElement('select');
        //Set attributes
        if(!['', undefined].includes(input.class)) field.classList.add(input.class);
        if(!['', undefined].includes(input.id)) field.setAttribute('id', input.id);
        if(!['', undefined].includes(input.placeholder)) field.setAttribute('placeholder', input.placeholder);
        if(!['', undefined].includes(input.name)) field.setAttribute('name', input.name);
        if(input.required === true) field.setAttribute('required', true);
        //Loop the options and add
        for(let option of input.options){
            let o = document.createElement('option');
            o.value = option.value;
            o.innerHTML = option.option;
            if(option.checked === true) o.setAttribute('selected', '');
            field.append(o);
        }
        //Return
        return field;
    }


    /* ------------------------------------------------------------------------------------------------------------------------
        Functions
    ------------------------------------------------------------------------------------------------------------------------ */


    /**
     * Bind next slide
     *
     * @param {Object} slideObj - The slide
     * @param {Int} key - The key
     * @return {Promise}
    */
    async nextSlideBtn(slideObj, key){
        //Try
        try {
            //Elems
            for(let btn of this.panda.querySelectorAll(`.slide[data-slide="${key}"] .footer-wrapper [data-action="next"]`)){
                btn.addEventListener('click', async () => {
                    //Get the current slide
                    let slide = btn.closest(`.slide[data-slide="${key}"]`),
                        nextNum = key + 1,
                        nextSlide = this.panda.querySelector(`.slide[data-slide="${nextNum}"]`),
                        inputs = slide.querySelectorAll('input[required], select[required], textarea[required]');
                    //Add a loader
                    slide.classList.add('loading');
                    //Do we have required fields that are empty?
                    let valid = true
                    //Loop the inputs
                    for(let input of inputs){
                        //Reset errors
                        input.removeAttribute('data-error');
                        //If required but empty, we dont need to type check
                        if(input.value == ''){
                            valid = false;
                        } else {
                            //Type check
                            switch(input.getAttribute('type')){
                                case 'email':
                                    valid = await validate.email(input.value);
                                break;
                                case 'tel':
                                    valid = await validate.telephone(input.value);
                                break;
                            }
                        }
                        //If not valid, mark it
                        if(!valid) input.setAttribute('data-error', true);
                    }
                    //Return if not valid
                    if(!valid) {
                        //Remove a loader
                        slide.classList.remove('loading');
                        return;
                    }
                    //Is there a prevalidate function?
                    if(slideObj.preValidate){
                        let preValidateResult = await slideObj.preValidate(slide);
                        if(preValidateResult !== true){
                            //Remove a loader
                            slide.classList.remove('loading');
                            return;
                        }
                    }
                    //Push our results
                    for(let input of inputs) this.results.inputs[input.getAttribute('name')] = input.value;
                    //Do we have a next slide to work with or are we finished?
                    if(nextSlide == undefined) return this.complete();
                    //Load the next slide
                    slide.classList.remove('active');
                    nextSlide.classList.add('active');
                    //Remove a loader
                    slide.classList.remove('loading');
                });
            }
        } catch(e){
            console.log('Error loading next slide', e);
        }
    }


    /**
     * Bind previous slide
     *
     * @param {Object} slideObj - The slide
     * @param {Int} key - The key
     * @return {Promise}
    */
    async previousSlideBtn(slideObj, key){
        //Try
        try {
            //Elems
            for(let btn of this.panda.querySelectorAll(`.slide[data-slide="${key}"] .footer-wrapper [data-action="previous"]`)){
                btn.addEventListener('click', async () => {
                    //Get the current slide
                    let slide = btn.closest('.slide'),
                        prevNum = key - 1,
                        prevSlide = this.panda.querySelector(`.slide[data-slide="${prevNum}"]`);
                    //Do we have a next slide to work with or are we finished?
                    if(prevSlide == undefined) return this.cancel();
                    //Load the next slide
                    slide.classList.remove('active');
                    prevSlide.classList.add('active');
                });
            }
        } catch(e){
            console.log('Error loading prev slide', e);
        }
    }


    /**
     * Bind custom slide btn
     *
     * @param {Object} slideObj - The slide
     * @param {Int} key - The key
     * @return {Promise}
    */
    async customSlideBtn(slideObj, key){
        //Try
        try {
            //Get panda instance
            let panda = this;
            //Loop the slide obj buttons and find our custom one
            for(let button of slideObj.buttons){
                //If custom action is a function
                if(button.action == 'custom' && button.customAction != undefined){
                    //Get matching btn
                    let btn = this.panda.querySelector(`.slide[data-slide="${key}"] .footer-wrapper [data-btn-id="${button.id}"]`);
                    //On click
                    btn.addEventListener('click', async () => {
                        //Run
                        button.customAction(slideObj, panda);
                    });
                }
            }
        } catch(e){
            console.log('Error loading custom slide btn', e);
        }
    }


    /**
     * Bind cancel btn
     *
     * @return {Promise}
    */
     async cancelBtn(){
        //Try
        try {
            //Elems
            let btn = this.panda.querySelector('.container > span.close');
            btn.addEventListener('click', () => this.cancel());
        } catch(e){
            console.log('Error cancel btn', e);
        }
    }


    /**
     * Cancel action
     *
     * @return {Promise}
    */
    async cancel(){
        //Populate the results
        this.results.complete = false;
        //Destroy it, its cancelled
        this.panda.remove();
        //Resolve
        return this.resolver(this.results);
    }


    /**
     * Close
     *
     * @return {Promise}
    */
    async close(){
        //Populate the results
        this.results.complete = false;
        this.results.inputs = [];
        //Destroy it, its cancelled
        this.panda.remove();
        //Resolve
        return;
    }


    /**
     * Complete
     *
     * @return {Promise}
    */
    async complete(){
        //Populate the results
        this.results.complete = true;
        //Close it
        this.panda.remove();
        //Resolve
        return this.resolver(this.results);
    }


    /**
     * Get the icons content
     *
     * @return {Promise}
    */
     async getIconContent(type){
        //Switch the type
        switch(type){
            case 'close':
                return '{{ICON_CLOSE}}';
            break;
        }
    }


    /* ------------------------------------------------------------------------------------------------------------------------
        Bindings
    ------------------------------------------------------------------------------------------------------------------------ */


    /**
     * Add all of the bindings for the panda
     *
     * @return {Promise}
    */
    async bindings(){
        //Let promises
        let promises = [];
        //Close btn?
        if(this.options.closeBtn){
            promises.push(this.cancelBtn());
        }
        //Loop the slides and bind separately
        for(let [key, slide] of Object.entries(this.slides)){
            //Correct the key
            key = parseInt(key) + 1;
            promises.push(this.nextSlideBtn(slide, key)),
            promises.push(this.previousSlideBtn(slide, key)),
            promises.push(this.customSlideBtn(slide, key))
        }
        //Return
        return await Promise.all(promises);
    }


}





/**
 * Export the panda function to use
 *
 * @param {Object} slides - The slides
 * @param {Object} options - The options
 * @returns {Promise}  - An instance of panda
 */
export const Panda = async (slides, options) => {
    //Start a new Panda
    let newBuild = new Builder(slides, options);
    //Run the build
    return await newBuild.create();
}